﻿using System;
using System.IO;
using System.Reflection;
using System.ServiceModel;
using System.ServiceModel.Description;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Xsl;

namespace KWatkins.NUnitTfs
{
    internal abstract class Publisher
    {
        protected Publisher(Options options)
        {
            Options = options;
        }

        internal void Publish()
        {
            // TODO: We could change this to accept a .trx on the command line, thus allowing people
            // to upload any .trx they like.
            var trx = TransformNUnitOutput();

            Console.WriteLine("Adding build information to .trx file...");
            AddBuildInformationToTrx(trx);

            if (Options.ConvertOnly)
            {
                Console.WriteLine("Saving .trx file...");
                trx.Save(Options.OutputFile);
                return;
            }

            Console.WriteLine("Publishing .trx file...");
            PublishToTfs(trx);
        }

        protected abstract void PublishToTfs(XDocument trx);

        private XDocument TransformNUnitOutput()
        {
            var xsl = LoadXsl();

            Console.WriteLine("Loading NUnit Output from {0}...", Options.NUnitOutput);
            using (var inputStream = File.OpenRead(Options.NUnitOutput))
            {
                using (var xmlReader = XmlReader.Create(inputStream))
                {
                    using (var outputStream = new MemoryStream())
                    {
                        using (var xmlWriter = XmlWriter.Create(outputStream))
                        {
                            Console.WriteLine("Transforming NUnit output to .trx file...");
                            xsl.Transform(xmlReader, xmlWriter);
                        }

                        outputStream.Flush();
                        outputStream.Position = 0;

                        using (var outputReader = XmlReader.Create(outputStream))
                        {
                            return XDocument.Load(outputReader);
                        }
                    }
                }
            }
        }

        private XslCompiledTransform LoadXsl()
        {
            var xsl = new XslCompiledTransform(true);

            // Has an XSLT been specified on the command line?
            if (!string.IsNullOrEmpty(Options.NUnitToMSTestXSLT))
            {
                // Yes, use it.
                xsl.Load(Options.NUnitToMSTestXSLT);
            }
            else
            {
                // No, take the one from the assembly. Use this Publisher's type to scope the XSLT; let's us have 
                // a different one in each TFS folder.
                using (var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(GetType(), "NUnitToMSTest.xslt"))
                {
                    using (var reader = XmlReader.Create(stream))
                    {
                        xsl.Load(reader);
                    }
                }
            }

            return xsl;
        }

        private void AddBuildInformationToTrx(XDocument document)
        {
            var timesName = XName.Get("Times", document.Root.Name.Namespace.NamespaceName);
            var timesElement = document.Root.Element(timesName);

            var buildName = XName.Get("Build", document.Root.Name.Namespace.NamespaceName);
            var buildElement =
                new XElement(buildName,
                    new XAttribute("flavor", Options.Flavour),
                    new XAttribute("platform", Options.Platform));

            timesElement.AddAfterSelf(buildElement);
        }

        protected static Guid GetRunId(XDocument trx)
        {
            return new Guid(trx.Root.Attribute("id").Value);
        }

        protected Options Options
        {
            get;
            private set;
        }

        protected void SetEndpointAddress(ServiceEndpoint endpoint, string uriSuffix)
        {
            if (string.IsNullOrEmpty(Options.ServerUri)) return;
            var endpointUri = Options.ServerUri;
            if (!endpointUri.EndsWith("/"))
            {
                endpointUri += "/";
            }
            endpointUri += uriSuffix;
            endpoint.Address = new EndpointAddress(endpointUri);
        }
    }
}
